from PIL import Image

import cv2


def resize_scaling(image_filename, width, height, output_filename = None):
    """
    Resizes the provided 'image_filename' to the provided 'width' and 'height' keeping the
    aspect ratio. This method enlarges the image to fit the desired size and then makes a 
    crop to obtain that size from the center of the resized image. If 'output_filename' is
    provided, the image is saved locally with that name.
    """
    image = Image.open(image_filename)
    image_width, image_height = image.size

    if image_width == width and image_height == height:
        return image.save(output_filename)

    aspect_ratio = image_width / image_height
    if aspect_ratio > (width / height):
        # Image is very horizontal, so width changes faster, we need to focus on height
        factor = (height * 100 / image_height) / 100
        image_width = int(image_width * factor)
        image_height = height
    else:
        # Image is very vertical, so height changes faster, we need to focus on width
        factor = (width * 100 / image_width) / 100
        image_width = 1920
        image_height = int(image_height * factor)
    image = image.resize((image_width, image_height))

    # We will crop form the center to edges
    left = 0
    right = width
    top = 0
    bottom = height
    if image_width > width:
        # If it is 1960 => leave [0, 20], get [20, 1940], leave [1940, 1960]
        margin = int((image_width - width) / 2)
        left = 0 + margin
        right = image_width - margin
        # We make and adjustment if some pixel left
        while (right - left) > width:
            right -= 1
        while (right - left) < width:
            if left > 0:
                left -= 1
            else:
                right += 1
    if image_height > height:
        # If it is 1140 => leave [0, 30], get [30, 1110], leave [1110, 1140]
        margin = int((image_height - height) / 2)
        top = 0 + margin
        bottom = image_height - margin
        # We make and adjustment if some pixel left
        while (bottom - top) > height:
            bottom -= 1
        while (bottom - top) < height:
            if top > 0:
                top -= 1
            else:
                bottom += 1

    image = image.crop((left, top, right, bottom))
    # Image that is 1920x1080 and is the center of the original image
    if output_filename:
        image.save(output_filename)

    return image

def resize_without_scaling(image_filename, width = 1920, height = 1080):
    """
    This method gets an image, resizes it and overwrites the original one.

    TODO: This method need work.
    """
    # TODO: We resize it simply, we don't care about scale
    image = cv2.imread(image_filename)
    resized_image = cv2.resize(image, dsize = (width, height), interpolation = cv2.INTER_CUBIC)
    cv2.imwrite(image_filename, resized_image)



from yta_general_utils.image.checker import is_valid_image
from yta_general_utils.file.filename import filename_is_type, FileType
from PIL import Image


# TODO: Apply types (str, Image.Image, etc.)
def recognize_image(image):
    """
    Recognizes the provided 'image' as an Image and returns it as
    it is or as a PIL Image if valid image filename is provided.
    """
    if not image:
        raise Exception('No "image" provided.')
    
    if isinstance(image, str):
        if not filename_is_type(image, FileType.IMAGE):
            raise Exception('The "image" parameter is not a valid image filename.')

        if not is_valid_image(image):
            raise Exception('The "image" parameter provided is not a valid image.')
        
        image = Image.open(image)

    # TODO: How to know if numpy is an image (?) It is a numpy,
    # yes, but is it an image?
    # if isinstance(image, np.ndarray):
    # TODO: Check if it is not a valid numpy image or base64 image or
    # anything else that can be an Image
    return image



from yta_general_utils.temp import create_temp_filename
from yta_general_utils.file.filename import get_file_extension, replace_file_extension
from pydub import AudioSegment


def mp3_to_wav(mp3_filename: str, output_filename: str):
    """
    Receives an .mp3 file 'mp3_filename' and turns it into a .wav
    file stored in 'output_filename'. This method will return the
    actual 'output_filename' used to export the file as it could
    has been changed by the method.
    """
    if not mp3_filename:
        raise Exception('No "mp3_filename" provided.')
    
    # TODO: Check that 'mp3_filename' is a valid mp3 filename and
    # that the file exists
    
    if not output_filename:
        # TODO: Replace this when not exporting needed
        output_filename = create_temp_filename('tmp_wav_sound.wav')

    # TODO: Make mp3 being not only a filename
    if not get_file_extension(output_filename) == 'wav':
        output_filename = replace_file_extension(output_filename, 'wav')

    # TODO: Generalize this method to swap between formats
    sound = AudioSegment.from_mp3(mp3_filename)
    # TODO: Achieve this without exporting, please
    sound.export(output_filename, format = "wav")

    return output_filename


from yta_general_utils.programming.error_message import ErrorMessage
from yta_general_utils.file.filename import filename_is_type, FileType
from yta_general_utils.checker.type import variable_is_positive_number
from yta_general_utils.file.checker import file_exists
from yta_general_utils.checker.url import url_is_ok
from enum import Enum
from typing import Union


class ParameterValidator:
    @classmethod
    def validate_mandatory_parameter(cls, name: str, value):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' has a value, raising an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not value:
            raise Exception(ErrorMessage.parameter_not_provided(name))

        return value
        
    @classmethod
    def validate_string_parameter(cls, name: str, value: str):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a string value, raising an Exception if
        not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not isinstance(value, str):
            raise Exception(ErrorMessage.parameter_is_not_string(name))

        return value

    @classmethod
    def validate_bool_parameter(cls, name: str, value: bool):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a boolean value, raising and Exception 
        if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not isinstance(value, bool):
            raise Exception(ErrorMessage.parameter_is_not_boolean(name))

        return value
        
    @classmethod
    def validate_file_exists(cls, name: str, value: str):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a file that actually exists, raising
        an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not file_exists(value):
            raise Exception(ErrorMessage.parameter_is_file_that_doesnt_exist(name))

        return value
        
    @classmethod
    def validate_filename_is_type(cls, name: str, value: str, file_type: FileType = FileType.IMAGE):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a filename of the given 'file_type',
        raising an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not filename_is_type(value, file_type):
            raise Exception(ErrorMessage.parameter_is_not_file_of_file_type(name, file_type))

        return value
        
    @classmethod
    def validate_url_is_ok(cls, name: str, value: str):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a valid url (the url is accessible),
        raising an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not url_is_ok(value):
            raise Exception(ErrorMessage.parameter_is_not_valid_url(name))

        return value
        
    @classmethod
    def validate_positive_number(cls, name: str, value: Union[int, float]):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a positive number (0 or greater),
        raising an Exception if not.

        This method returns the provided 'value' as it is if 
        everything is ok.
        """
        if not variable_is_positive_number(value):
            raise Exception(ErrorMessage.parameter_is_not_positive_number(name))

        return value

    @classmethod
    def validate_is_class(cls, name: str, value, class_names: Union[list[str], str]):
        """
        Validates if the provided 'value' is one of the provided 'class_names'
        by using the 'type(value).__name__' checking.

        This method returns the 'value' as it is if everything is ok.
        """
        if isinstance(class_names, str):
            class_names = [class_names]

        if not type(value).__name__ in class_names:
            raise Exception(ErrorMessage.parameter_is_not_class(name, class_names))
        
        return value
        
    # Complex ones below
    @classmethod
    def validate_string_mandatory_parameter(cls, name: str, value: str):
        """
        Validates if the provided 'value' is a valid and non
        empty string.
        """
        cls.validate_mandatory_parameter(name, value)
        cls.validate_string_parameter(name, value)

        return value

    @classmethod
    def validate_numeric_positive_mandatory_parameter(cls, name: str, value: str):
        """
        Validates if the provided 'value' is a positive numeric
        value.
        """
        cls.validate_mandatory_parameter(name, value)
        cls.validate_positive_number(name, value)

        return value
    
    @classmethod
    def validate_is_enum(cls, enum: Union['YTAEnum', Enum]):
        """
        Validates if the provided 'value' is a valid Enum
        class or subclass.

        This method will raise an Exception if the provided
        'value' is not a valid Enum class or subclass, or
        will return it as it is if yes.
        """
        if not isinstance(enum, Enum) and not issubclass(enum, Enum):
            raise Exception(f'The parameter "{enum}" provided is not an Enum class or subclass.')
        
        return enum

    @classmethod
    def validate_enum(cls, name: str, value: Union['YTAEnum', str], enum: 'YTAEnum'):
        """
        Validates if the provided 'value' value is a valid
        Enum or Enum value of the also provided 'enum' class.
        The 'name' parameter is used to give a more specific
        error message.

        This method will raise an Exception if something is
        wrong or will return the 'value' as an 'enum' Enum.
        instance if everything is ok.
        """
        cls.validate_mandatory_parameter(name, value)
        cls.validate_is_class(name, value, [enum.__class__, 'str'])

        if isinstance(value, str):
            if not enum.is_valid(value):
                raise Exception(f'The "{name}" parameter provided is not a valid {enum} value.')
            
            value = enum(value)

        return value


from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from yta_general_utils.web.scrapper.chrome_scrapper import go_to_and_wait_loaded, start_chrome

import time
import os
import pyautogui
import glob
import shutil

MESSAGE_SENDER_ME = 1
MESSAGE_SENDER_OTHER = 2
PROJECT_ABSOLUTE_PATH = os.getenv('PROJECT_ABSOLUTE_PATH')
TOUSE_ABSOLUTE_PATH = os.getenv('TOUSE_ABSOLUTE_PATH')
# TODO: Move this to .env variable
DOWNLOADS_ABSOLUTE_PATH = 'C:/Users/dania/Downloads/'

def __setup(driver, name, conversation_date, profile_picture_absolute_path):
    # TODO: Set maximum 'name' and 'username' lengths
    # TODO: Implement 'status' (?)
    __set_name(driver, name)
    __set_status(driver, 'En línea')
    __set_conversation_date(driver, conversation_date)
    __set_profile_picture(driver, profile_picture_absolute_path)

def __set_profile_picture(driver, absolute_path):
    element_present = EC.presence_of_element_located((By.ID, 'profileImg'))
    element = WebDriverWait(driver, 10).until(element_present)
    time.sleep(1)
    element = element.find_element(By.XPATH, '..')
    element.click()
    time.sleep(2)
    pyautogui.write((absolute_path).replace('/', '\\')) 
    time.sleep(1)
    pyautogui.press('enter')
    time.sleep(1)

def __set_conversation_date(driver, str):
    date_input = driver.find_element(By.ID, 'edit-chatday')
    time.sleep(1)
    date_input.clear()
    date_input.send_keys(str)

def __send_my_photo(driver, photo_absolute_path, hour):
    driver.execute_script('window.scrollTo(0, 0)')
    profile2_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person2")]')
    profile2_tab.click()
    element_present = EC.presence_of_element_located((By.ID, 'SecondImage'))
    element = WebDriverWait(driver, 10).until(element_present)
    time.sleep(1)
    element = element.find_element(By.XPATH, '..')
    element.click()
    time.sleep(2)
    pyautogui.write((photo_absolute_path).replace('/', '\\')) 
    time.sleep(1)
    pyautogui.press('enter')
    time.sleep(1)
    profile2_time_input = driver.find_element(By.ID, 'time-profile2')
    profile2_time_input.clear()
    profile2_time_input.send_keys(hour)
    time.sleep(1)
    # Set message as read
    profile2_message_is_read_radiobutton = driver.find_element(By.ID, 'read')
    if not profile2_message_is_read_radiobutton.is_selected():
        profile2_message_is_read_radiobutton.click()
    time.sleep(1)
    profile2_textarea = driver.find_element(By.ID, 'profile2_message')
    profile2_textarea.clear()
    profile2_textarea.send_keys('')
    time.sleep(1)
    actions = ActionChains(driver)
    for i in range(2):
        actions.send_keys(Keys.TAB)
    actions.send_keys(Keys.SPACE)
    actions.perform()
    time.sleep(1)

def __send_other_photo(driver, photo_absolute_path, hour):
    driver.execute_script('window.scrollTo(0, 0)')
    profile2_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person1")]')
    profile2_tab.click()
    element_present = EC.presence_of_element_located((By.ID, 'InputFirstImage'))
    element = WebDriverWait(driver, 10).until(element_present)
    time.sleep(1)
    element = element.find_element(By.XPATH, '..')
    element.click()
    time.sleep(2)
    pyautogui.write((photo_absolute_path).replace('/', '\\')) 
    time.sleep(1)
    pyautogui.press('enter')
    time.sleep(1)
    profile1_time_input = driver.find_element(By.ID, 'time-profile1')
    profile1_time_input.clear()
    profile1_time_input.send_keys(hour)
    time.sleep(1)
    profile1_textarea = driver.find_element(By.ID, 'profile1_message')
    profile1_textarea.clear()
    profile1_textarea.send_keys('')
    time.sleep(1)
    actions = ActionChains(driver)
    actions.send_keys(Keys.TAB)
    actions.send_keys(Keys.SPACE)
    actions.perform()
    time.sleep(1)

def __set_name(driver, name):
    profile1_name_input = driver.find_element(By.ID, 'profile1_name')
    time.sleep(1)
    profile1_name_input.clear()
    profile1_name_input.send_keys(name)

def __set_status(driver, status):
    status_input = driver.find_element(By.ID, 'edit-Status')
    time.sleep(1)
    status_input.clear()
    status_input.send_keys(status)

def __download(driver, output_filename):
    time.sleep(2)
    download_button = driver.find_element(By.ID, 'snapshot')
    download_button.click()
    time.sleep(10)
    # Move downloaded file to our destination
    list_of_files = glob.glob(DOWNLOADS_ABSOLUTE_PATH + '*.png')
    latest_file = max(list_of_files, key = os.path.getctime)
    shutil.move(latest_file, os.path.join(PROJECT_ABSOLUTE_PATH + output_filename))

def __write_my_message(driver, message, hour):
    driver.execute_script('window.scrollTo(0, 0)')
    profile2_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person2")]')
    profile2_tab.click()
    time.sleep(1)
    profile2_time_input = driver.find_element(By.ID, 'time-profile2')
    profile2_time_input.clear()
    profile2_time_input.send_keys(hour)
    time.sleep(1)
    # Set message as read
    profile2_message_is_read_radiobutton = driver.find_element(By.ID, 'read')
    if not profile2_message_is_read_radiobutton.is_selected():
        profile2_message_is_read_radiobutton.click()
    time.sleep(1)
    profile2_textarea = driver.find_element(By.ID, 'profile2_message')
    profile2_textarea.clear()
    profile2_textarea.send_keys(message)
    time.sleep(1)
    actions = ActionChains(driver)
    for i in range(2):
        actions.send_keys(Keys.TAB)
    actions.send_keys(Keys.SPACE)
    actions.perform()
    time.sleep(1)

def __write_other_message(driver, message, hour):
    driver.execute_script('window.scrollTo(0, 0)')
    profile1_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person1")]')
    profile1_tab.click()
    time.sleep(1)
    profile1_time_input = driver.find_element(By.ID, 'time-profile1')
    profile1_time_input.clear()
    profile1_time_input.send_keys(hour)
    time.sleep(1)
    profile1_textarea = driver.find_element(By.ID, 'profile1_message')
    profile1_textarea.clear()
    profile1_textarea.send_keys(message)
    time.sleep(1)
    actions = ActionChains(driver)
    actions.send_keys(Keys.TAB)
    actions.send_keys(Keys.SPACE)
    actions.perform()
    time.sleep(1)

# TODO: Implement sending foto
    
def generate_conversation(name, messages, conversation_date, profile_picture_absolute_path, output_filename):
    # Messages must come with autor and text
    URL = 'https://prankshit.com/fake-whatsapp-chat-generator.php'

    profile_picture_absolute_path = profile_picture_absolute_path.replace('/', '\\')

    try:
        driver = start_chrome(True)
        go_to_and_wait_loaded(driver, URL)

        # I force 'conversation_date' by now
        conversation_date = 'Hoy'

        time.sleep(2)
        __setup(driver, name, conversation_date, profile_picture_absolute_path)
        time.sleep(1)

        # Add more data ('active')
        messages = [
            {
                'sender': MESSAGE_SENDER_ME,
                'message': 'Holaaaaa!',
                'hour': '22:07',
            },
            {
                'sender': MESSAGE_SENDER_ME,
                'message': '',
                'photo': 'C:/Users/dania/Desktop/PROYECTOS/Youtube/stockpediayt/software/ai.png',
                'hour': '22:07',
            },
            {
                'sender': MESSAGE_SENDER_ME,
                'message': 'estás??',
                'hour': '22:07',
            },
            {
                'sender': MESSAGE_SENDER_OTHER,
                'message': 'Sí, ¿qué pasa wacho??',
                'hour': '22:09',
            },
            {
                'sender': MESSAGE_SENDER_OTHER,
                'message': '',
                'photo': 'C:/Users/dania/Desktop/PROYECTOS/Youtube/stockpediayt/software/ai_image.png',
                'hour': '22:10',
            },
            {
                'sender': MESSAGE_SENDER_ME,
                'message': 'Nada, solo quería saludarte!! Campeón!',
                'hour': '22:10',
            },
            {
                'sender': MESSAGE_SENDER_OTHER,
                'message': 'Dalee! Grande, gracias!!',
                'hour': '22:11'
            },
        ]

        for index, message in enumerate(messages):
            if message['sender'] == MESSAGE_SENDER_ME:
                if 'photo' in message and message['photo']:
                    __send_my_photo(driver, message['photo'], message['hour'])
                else:
                    __write_my_message(driver, message['message'], message['hour'])
            else:
                if 'photo' in message and message['photo']:
                    __send_other_photo(driver, message['photo'], message['hour'])
                else:
                    __write_other_message(driver, message['message'], message['hour'])

            __download(driver, 'wip/test_whatsapp_' + str(index) + '.png')
    finally:
        driver.close()


from yta_general_utils.file.enums import FileSearchOption
from requests import Response

import json
import os


def write_binary_file(binary_data: bytes, filename: str):
    """
    Writes the provided 'binary_data' in the 'filename' file. It replaces the
    previous content if existing.
    """
    if not binary_data:
        return None
    
    if not filename:
        return None
    
    f = open(filename, 'wb')
    f.write(binary_data)
    f.close()

def write_json_to_file(dict: dict, filename: str):
    """
    Writes the provided 'dict' as a json into the 'filename'.

    @param
        **dict**
        Python dictionary that will be stored as a json.

        **filename**
        File path in which we are going to store the information.
    """
    if dict == None:
        return None
    
    if not filename:
        return None
    
    return write_file(json.dumps(dict, indent = 4), filename)

def write_file(text: str, output_filename: str):
    """
    Writes the provided 'text' in the 'filename' file. It replaces the previous content
    if existing.
    """
    if not text:
        return None
    
    if not output_filename:
        return None

    f = open(output_filename, 'w', encoding = 'utf8')
    f.write(text)
    f.close()

    return output_filename

def write_file_by_chunks_from_response(response: Response, output_filename: str):
    """
    Iterates over the provided 'response' and writes its content
    chunk by chunk in the also provided 'output_filename'.

    TODO: If you find a better way to handle this you are free to
    create new methods and move them into a new file.
    """
    if not response:
        return None
    
    if not output_filename:
        return None
    
    CHUNK_SIZE = 32768

    # TODO: Make this method work with a common Iterator parameter
    # and not an specific response, please
    with open(output_filename, 'wb') as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)

    return output_filename

def delete_files(folder, pattern = '*'):
    """
    Delete all the files in the 'folder' provided that match the provided
    'pattern'. The default pattern removes all existing files, so please
    use this method carefully.
    """
    # TODO: Make some risky checkings  about removing '/', '/home', etc.
    files = list(folder, FileSearchOption.FILES_ONLY, pattern)
    # TODO: Check what happens if deleting folders with files inside
    for file in files:
        os.remove(file)


import unicodedata
import re


def remove_accents(text):
    """
    Removes the accents from the provided 'text'.
    """
    return ''.join(char for char in unicodedata.normalize('NFD', text) if unicodedata.category(char) != 'Mn')

def remove_marks(text):
    """
    Removes the existing quotation and exclamation marks and also commas and full stops.

    # This below could work eliminating no white spaces.
    pattern = re.compile('[\W_]+')
    return pattern.sub('', s)
    """
    # TODO: Try with a 're' implementation
    return text.replace('?', '').replace('¿', '').replace(',', '').replace('.', '').replace('¡', '').replace('!', '').replace('(', '').replace(')', '')

def remove_marks_and_accents(text):
    """
    Removes the existing accents and quotation and exclamation marks, and commas and full stops.
    """
    return remove_accents(remove_marks(text))

def remove_non_ascii_characters(text):
    """
    Removes any non-ascii character from the provided 'text' and returns it
    modified.
    """
    s = list(remove_accents(text))
    index = 0
    while index < len(s):
        char = s[index]
        if not char.isascii():
            del s[index]
        else:
            index += 1

    return ''.join(s)

def fix_unseparated_periods(text: str):
    """
    This method fixes the provided 'text' by applying a space
    after any period without it.
    """
    # Thank you: https://stackoverflow.com/a/70394076
    text = re.sub(r'\.(?!(?<=\d\.)\d) ?', '. ', text)
    
    return text

def fix_separated_parenthesis(text: str):
    """
    This method remove the spaces that are found before or/and
    after any parenthesis. For example, '( esto es de prueba )'
    will become '(esto es de prueba)'.
    """
    def fix_match(match):
        opening, content, closure = match.groups()
        if content:
            # Remove the whitespaces
            without_whitespaces = re.sub(r'\s+', ' ', content.strip())
            return f'{opening}{without_whitespaces}{closure}'
        
        return match.group(0)
    
    text = re.sub(r'(\()([^\(\)]+?)(\))', fix_match, text)
    
    return text

def fix_separated_square_brackets(text: str):
    """
    This method removes the spaces that are found before or/and
    after any square bracket. For example, '[  shortcode ]' will
    become '[shortcode]'
    """
    def fix_match(match):
        opening, content, closure = match.groups()
        if content:
            # Remove the whitespaces
            without_whitespaces = re.sub(r'\s+', ' ', content.strip())
            return f'{opening}{without_whitespaces}{closure}'
        
        return match.group(0)
    
    text = re.sub(r'(\[)([^\[\]]+?)(\])', fix_match, text)
    
    return text

def add_missing_spaces_before_and_after_parenthesis(text: str):
    """
    This method adds the missing spaces before and after any 
    parenthesis in the provided 'text'. For example, 'hola(nueve)siete'
    will become 'hola (nueve) siete'.
    """
    text = re.sub(r'(?<!\s)\((?!\s)', ' (', text)
    text = re.sub(r'\)(?!\s)', ') ', text)
    
    return text

def add_missing_spaces_before_and_after_square_brackets(text: str):
    """
    This method adds the missing spaces before and after any 
    square bracket in the provided 'text'. For example, 
    'hola[nueve]siete' will become 'hola [nueve] siete'.
    """
    text = re.sub(r'(?<!\s)\[(?!\s)', ' [', text)
    text = re.sub(r'\](?!\s)', '] ', text)
    
    return text

def fix_excesive_blank_spaces(text: str):
    """
    Checks the provided 'text' and removes the extra blank spaces.
    This means that any sentence with more than one blank space 
    will be replaced by only one blank space.
    """
    # TODO: Ok, why am I using not the 'repl' param in re.search?
    # I'm applying it in the new method below, please check if
    # valid to avoid the while, thank you
    filtered = re.search('[ ]{2,}', text)
    while filtered:
        index_to_replace = filtered.end() - 1
        s = list(text)
        s[index_to_replace] = ''
        text = ''.join(s)
        filtered = re.search('[ ]{2,}', text)

    return text

def numbers_to_text(text):
    """
    This method receives a text that could contain numbers and turns
    those numbers into text.

    This method is useful to let narration software work with just
    text and avoid numbers problems.
    """
    words = str(text).split(' ')

    SPECIAL_CHARS = ['¡', '!', ',', '.', '¿', '?', ':', '"', '\'', '#', '@']
    new_words = []
    # Iterate over each word to turn numbers into words
    for word in words:
        begining = ''
        ending = ''

        # We need to remove special chars at the begining or at the ending
        # to be able to work well with the important part of the word, but
        # we cannot simply delete ',' or '.' because could be in the middle
        # of a word
        if word[0] in SPECIAL_CHARS:
            begining = word[0]
            word = word[1:]
        if word[len(word) - 1] in SPECIAL_CHARS:
            ending = word[len(word) - 1]
            word = word[:1]

        try:
            word = float(word)
            # If here, it is a number, lets change its name
            # TODO: Implement logic here, so word will be the text, not the number
            print('Processing number: ' + str(word))
            accumulated_text = ''
            # We receive 123.456.789
            is_million = False
            is_one = False
            is_thousand = False
            is_ten = False
            divisor = 1000000000
            res = int(word / divisor)  # 1 . 000 . 000 . 000
            if res >= 1:
                is_million = True
                is_thousand = True
                accumulated_text += __process_unit(res)
                word -= divisor * res

            if is_thousand:
                accumulated_text += ' mil'
                is_thousand = False

            divisor = 100000000
            res = int(word / divisor)  # 100 . 000 . 000
            if res >= 1:
                is_million = True
                if res == 1:
                    is_one = True
                accumulated_text += __process_hundred(res)
                word -= divisor * res

            divisor = 10000000
            res = int(word / divisor) # 10 . 000 . 000
            if res >= 1:
                is_million = True
                is_ten = True
                if is_one:
                    accumulated_text += 'to'
                    is_one = False
                accumulated_text += __process_ten(res)
                word -= divisor * res

            divisor = 1000000
            res = int(word / divisor) # 1 . 000 . 000
            if res >= 1:
                is_million = True
                if is_one:
                    accumulated_text += 'to'
                    is_one: False
                if is_ten:
                    accumulated_text += ' y '
                    is_ten = False
                accumulated_text += __process_unit(res)
                word -= divisor * res

            if is_million:
                accumulated_text += ' millones'
                is_million = False

            divisor = 100000
            res = int(word / divisor) # 100 . 000
            if res >= 1:
                is_thousand = True
                if res == 1:
                    is_one = True
                accumulated_text += __process_hundred(res)
                word -= divisor * res

            divisor = 10000
            res = int(word / divisor) # 10 . 000
            if res >= 1:
                is_thousand = True
                is_ten = True
                if is_one:
                    accumulated_text += 'to'
                    is_one = False
                accumulated_text += __process_ten(res)
                word -= divisor * res

            divisor = 1000
            res = int(word / divisor) # 1 . 000
            if res >= 1:
                is_thousand = True
                if is_one:
                    accumulated_text += 'to'
                    is_one = False
                if is_ten:
                    accumulated_text += ' y '
                    is_ten = False
                accumulated_text += __process_unit(res)
                word -= divisor * res

            if is_thousand:
                accumulated_text += ' mil'
                is_thousand = False

            divisor = 100
            res = int(word / divisor) # 100
            if res >= 1:
                is_thousand = True
                if res == 1:
                    is_one = True
                accumulated_text += __process_hundred(res)
                word -= divisor * res

            divisor = 10
            res = int(word / divisor) # 10
            if res >= 1:
                is_thousand = True
                is_ten = True
                if is_one:
                    accumulated_text += 'to'
                    is_one = False
                accumulated_text += __process_ten(res)
                word -= divisor * res

            divisor = 1
            res = int(word / divisor) # 1
            if res >= 1:
                is_thousand = True
                if is_one:
                    accumulated_text += 'to'
                    is_one = False
                if is_ten:
                    accumulated_text += ' y '
                    is_ten = False
                accumulated_text += __process_unit(res)
                word -= divisor * res

            accumulated_text = accumulated_text.replace('  ', ' ').strip()
            # We need to replace in special cases
            accumulated_text = accumulated_text.replace('veinte y nueve', 'veintinueve')
            accumulated_text = accumulated_text.replace('veinte y ocho', 'veintiocho')
            accumulated_text = accumulated_text.replace('veinte y siete', 'veintisiete')
            accumulated_text = accumulated_text.replace('veinte y seis', 'veintiséis')
            accumulated_text = accumulated_text.replace('veinte y cinco', 'veinticinco')
            accumulated_text = accumulated_text.replace('veinte y cuatro', 'veinticuatro')
            accumulated_text = accumulated_text.replace('veinte y tres', 'veintitrés')
            accumulated_text = accumulated_text.replace('veinte y dos', 'veintidós')
            accumulated_text = accumulated_text.replace('veinte y uno', 'veintiuno')
            accumulated_text = accumulated_text.replace('diez y nueve', 'diecinueve')
            accumulated_text = accumulated_text.replace('diez y ocho', 'dieciocho')
            accumulated_text = accumulated_text.replace('diez y siete', 'diecisiete')
            accumulated_text = accumulated_text.replace('diez y seis', 'dieciséis')
            accumulated_text = accumulated_text.replace('diez y cinco', 'quince')
            accumulated_text = accumulated_text.replace('diez y cuatro', 'catorce')
            accumulated_text = accumulated_text.replace('diez y tres', 'trece')
            accumulated_text = accumulated_text.replace('diez y dos', 'doce')
            accumulated_text = accumulated_text.replace('diez y uno', 'once')

            word = accumulated_text
        except:
            pass

        new_words.append(begining + str(word) + ending)

    # We have the same size in 'words' and 'new_words', so lets build it
    final_text = " ".join(new_words)
    
    return final_text


def __process_hundred(number):
    """
    Receives a digit that represente the X00 part of a number and returns
    that number (according to its position) in words.

    This method returns ' novecientos' for 9 input and ' cien' for 1 input.
    """
    number_in_text = ''

    if number == 9:
        number_in_text = ' novecientos'
    elif number == 8:
        number_in_text = ' ochocientos'
    elif number == 7:
        number_in_text = ' setecientos'
    elif number == 6:
        number_in_text = ' seiscientos'
    elif number == 5:
        number_in_text = ' quinientos'
    elif number == 4:
        number_in_text = ' cuatrocientos'
    elif number == 3:
        number_in_text = ' trescientos'
    elif number == 2:
        number_in_text = ' doscientos'
    elif number == 1:
        number_in_text = ' cien'

    return number_in_text

def __process_ten(number):
    """
    Receives a digit that represente the X0 part of a number and returns
    that number (according to its position) in words.

    This method returns ' noventa' for 9 input and ' diez' for 1 input.
    """
    number_in_text = ''

    if number == 9:
        number_in_text = ' noventa'
    elif number == 8:
        number_in_text = ' ochenta'
    elif number == 7:
        number_in_text = ' setenta'
    elif number == 6:
        number_in_text = ' sesenta'
    elif number == 5:
        number_in_text = ' cincuenta'
    elif number == 4:
        number_in_text = ' cuarenta'
    elif number == 3:
        number_in_text = ' treinta'
    elif number == 2:
        number_in_text = ' veinte'
    elif number == 1:
        number_in_text = ' diez'

    return number_in_text

def __process_unit(number):
    """
    Receives a digit that represente the X part of a number and returns
    that number (according to its position) in words.

    This method returns ' nueve' for 9 input, and ' uno' for 1 input.
    """
    number_in_text = ''

    if number == 9:
        number_in_text = ' nueve'
    elif number == 8:
        number_in_text = ' ocho'
    elif number == 7:
        number_in_text = ' siete'
    elif number == 6:
        number_in_text = ' seis'
    elif number == 5:
        number_in_text = ' cinco'
    elif number == 4:
        number_in_text = ' cuatro'
    elif number == 3:
        number_in_text = ' tres'
    elif number == 2:
        number_in_text = ' dos'
    elif number == 1:
        number_in_text = ' uno'

    return number_in_text


from typing import Union

import inspect


def variable_is_type(variable, type):
    """
    Checks if the type of the provided 'variable' is the also provided
    'type', returning True if yes or False if not.
    """
    # TODO: Maybe let this accept array of types to check if one of 
    # them (?)
    # Check this to validate 'type' is array to treat it like an array:
    # https://stackoverflow.com/a/16807050
    if isinstance(variable, type):
        return True
    
    return False

def variable_is_number(variable: Union[int, float, str]):
    """
    This method checks if the provided 'variable' is a numeric type,
    or tries to cast it as a float number if string provided, and
    returns True in the only case that the 'variable' is actual a
    number by itself or as a string.
    """
    if not isinstance(variable, (int, float, str)):
        return False
    
    if isinstance(variable, str):
        try:
            float(variable)
        except:
            return False
        
    return True

def variable_is_positive_number(variable: Union[int, float, str]):
    """
    This method checks if the provided 'variable' is a numeric type,
    or tries to cast it as a float number if string provided, and
    returns True in the only case that the 'variable' is actual a
    number by itself or as a string and it is 0 or above it.
    """
    if not isinstance(variable, (int, float, str)):
        return False
    
    if isinstance(variable, str):
        try:
            variable = float(variable)
        except:
            return False
        
    return variable >= 0

def code_file_is(parameter, filename):
    """
    Checks if the provided parameter code is contained in the also
    provided 'filename'. This method is useful to check Enum objects
    or classes as we know the name we use for the files.

    This method was created to be able to check if a function that
    was passed as parameter was part of a custom Enum we created
    and so we could validate the was actually that custom Enum.
    Working with classes was not returning the custom Enum class
    created, so we needed this.
    """
    if inspect.getfile(parameter).endswith(filename):
        return True
    
    return False



import requests


def __is_short_tiktok_url(url: str):
    # url like: https://vm.tiktok.com/ZGeSJ6YRA
    if not url:
        return False
    
    # TODO: Do this with a regexp
    if not '/' in url:
        return False
    
    url = __clean(url)
    
    if 'vm.tiktok.com' in url:
        return True
    
    return False

def __is_long_tiktok_url(url: str):
    # url like: https://www.tiktok.com/@ahorayasabesque/video/7327001175616703777?_t=8jqq93LWqsC&_r=1
    if not url:
        return False
    
    # TODO: Do this with a regexp
    if not '/' in url:
        return False
    
    url = __clean(url)

    if 'www.tiktok.com' in url:
        return True
    
    return False

def __short_tiktok_url_to_long_tiktok_url(url: str):
    """
    Transforms the provided short tiktok 'url' to 
    the long format and returns it.
    """
    if not url:
        raise Exception('No "url" provided.')
    
    if not __is_short_tiktok_url(url):
        raise Exception('No "url" provided is not a short tiktok url.')

    return requests.get(url).url

def __clean(url: str):
    """
    Removes any additional parameter that is after a
    question mark sign.
    """
    if not url:
        raise Exception('No "url" provided.')

    if '?' in url:
        url = url.split('?')[0]

    return url

def is_valid_tiktok_url(url: str):
    """
    This method returns True if the provided 'url' is a
    valid long or short tiktok video url or False if not.
    """
    return __is_short_tiktok_url(url) or __is_long_tiktok_url(url)

def parse_tiktok_url(url: str):
    """
    Parses the provided 'url' returning a dict with that
    'url' (long version) and also the 'username' and the
    'video_id' if it is a valid url. It will raise an 
    Exception if not valid 'url' provided.
    """
    if not url:
        raise Exception('No "url" provided.')

    if not is_valid_tiktok_url(url):
        raise Exception('The provided "url" is not a valid tiktok video url.')
    
    url = __clean(url)
    if not __is_long_tiktok_url(url):
        url = __short_tiktok_url_to_long_tiktok_url(url)

    aux = url.split('/')

    data = {
        'username': aux[len(aux) - 3],
        'video_id': aux[len(aux) - 1],
        'url': url,
    }

    return data
    

    


from yta_general_utils.image.color import get_most_common_rgb_color
from PIL import Image
from subprocess import run

import numpy as np
import skimage.exposure
import cv2


def get_green_screen_position(image_filename):
    """
    Detects the green screen color of the provided 'image_filename' and then looks for
    the upper left corner and the bottom right corner.

    This method return an object containing 'ulx', 'uly', 'drx', 'dry' coords. It also
    returns the 'rgb_color' most common green color as an (r, g, b).

    This will return None in ['rgb_color'] field if no green color detected.
    """
    green_rgb_color = get_most_common_rgb_color(image_filename)

    image = Image.open(image_filename).convert('RGB')

    upper_left = {
        'x': 99999,
        'y': 99999,
    }
    down_right = {
        'x': -1,
        'y': -1,
    }

    for x in range(image.width):
        for y in range(image.height):
            rgb_color = (r, g, b) = image.getpixel((x, y))

            if rgb_color == green_rgb_color:
                if x < upper_left['x']:
                    upper_left['x'] = x
                if y < upper_left['y']:
                    upper_left['y'] = y
                
                """
                if x <= upper_left['x'] and y <= upper_left['y']:
                    upper_left = {
                        'x': x,
                        'y': y,
                    }
                """

                if x > down_right['x']:
                    down_right['x'] = x
                if y > down_right['y']:
                    down_right['y'] = y

                """
                if x >= down_right['x'] and y >= down_right['y']:
                    down_right = {
                        'x': x,
                        'y': y,
                    }
                """

    # We apply some margin to make sure we fit the green screen
    MARGIN = 2

    if (upper_left['x'] - MARGIN) > 0:
        upper_left['x'] -= MARGIN
    else:
        upper_left['x'] = 0

    if (upper_left['y'] - MARGIN) > 0:
        upper_left['y'] -= MARGIN
    else:
        upper_left['y'] = 0

    if (down_right['x'] + MARGIN) < 1920:
        down_right['x'] += MARGIN
    else:
        down_right['x'] = 1920

    if (down_right['y'] + MARGIN) < 1080:
        down_right['y'] += MARGIN
    else:
        down_right['y'] = 1080

    return {
        'rgb_color': green_rgb_color,
        'ulx': upper_left['x'],
        'uly': upper_left['y'],
        'drx': down_right['x'],
        'dry': down_right['y'],
    }

def remove_background(image_filename, output_filename):
    """
    Removes the background of the provided 'image_filename' by using the 
    'backgroundremover' open library that is included in a comment.
    """
    # It uses (https://github.com/nadermx/backgroundremover?referral=top-free-background-removal-tools-apis-and-open-source-models)
    # That uses U2Net (https://medium.com/axinc-ai/u2net-a-machine-learning-model-that-performs-object-cropping-in-a-single-shot-48adfc158483)
    command_parameters = ['backgroundremover', '-i', image_filename, '-o', output_filename]

    run(command_parameters)

    # TODO: This below seems to work (as shown in this 
    # commit https://github.com/nadermx/backgroundremover/commit/c590858de4c7e75805af9b8ecdd22baf03a1368f)
    """
    from backgroundremover.bg import remove
    def remove_bg(src_img_path, out_img_path):
        model_choices = ["u2net", "u2net_human_seg", "u2netp"]
        f = open(src_img_path, "rb")
        data = f.read()
        img = remove(data, model_name=model_choices[0],
                    alpha_matting=True,
                    alpha_matting_foreground_threshold=240,
                    alpha_matting_background_threshold=10,
                    alpha_matting_erode_structure_size=10,
                    alpha_matting_base_size=1000)
        f.close()
        f = open(out_img_path, "wb")
        f.write(img)
        f.close()
    """

def remove_background_video(video_filename, output_filename):
    # TODO: Move to 'video_utils'
    # TODO: This is too demanding as I cannot process it properly
    # Output must end in .mov to preserve transparency
    command_parameters = ['backgroundremover', '-i', video_filename, '-tv', '-o', output_filename]

    run(command_parameters)

def remove_green_screen(image_filename, output_filename):
    # From https://stackoverflow.com/a/72280828
    # load image
    img = cv2.imread(image_filename)

    # convert to LAB
    lab = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)

    # extract A channel
    A = lab[:,:,1]

    # threshold A channel
    thresh = cv2.threshold(A, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]

    # blur threshold image
    blur = cv2.GaussianBlur(thresh, (0,0), sigmaX = 5, sigmaY = 5, borderType = cv2.BORDER_DEFAULT)

    # stretch so that 255 -> 255 and 127.5 -> 0
    mask = skimage.exposure.rescale_intensity(blur, in_range = (127.5, 255), out_range = (0, 255)).astype(np.uint8)

    # add mask to image as alpha channel
    result = img.copy()
    result = cv2.cvtColor(img,cv2.COLOR_BGR2BGRA)
    result[:,:,3] = mask

    # save output
    cv2.imwrite(output_filename, result)



from yta_general_utils.programming.parameter_validator import ParameterValidator
from yta_general_utils.programming.error_message import ErrorMessage
from enum import Enum


class YTAEnum(Enum):
    @classmethod
    def is_valid_value(cls, value):
        """
        This method returns True if the provided 'value' is a valid
        value for this YTAEnum class, or False if not.
        """
        return is_valid_value(value, cls)
    
    @classmethod
    def is_valid_name(cls, name):
        """
        This method returns True if the provided 'name' is a valid
        name of this YTAEnum class, or False if not.
        """
        return is_valid_name(name, cls)
    
    @classmethod
    def is_valid_name_or_value(cls, name_or_value):
        """
        This method returns True if the provided 'name_or_value' is
        a valid name or a valid value of this YTAEnum class, or 
        False if not.
        """
        return is_valid_name_or_value(name_or_value, cls)
    
    @classmethod
    def is_valid(cls, name_or_value_or_enum):
        """
        This method returns True if the provided 'name_or_value_or_enum'
        is a valid name, a valid value or a valid instance of this
        YTAEnum class, or returns False if not.
        """
        return is_valid(name_or_value_or_enum, cls)
    
    @classmethod
    def name_to_enum(cls, name):
        """
        This method returns this 'enum' YTAEnum item instance if the
        provided 'name' is a valid name of that 'enum' YTAEnum class,
        or raises an Exception if not.
        """
        return from_name(name, cls)
    
    @classmethod
    def value_to_enum(cls, value):
        """
        This method returns the 'enum' YTAEnum item instance if the
        provided 'value' is a valid value of that 'enum' YTAEnum class,
        or raises an Exception if not.
        """
        return from_value(value, cls)
    
    @classmethod
    def name_or_value_to_enum(cls, name_or_value):
        """
        This method returns this 'enum' YTAEnum item instance if the
        provided 'name_or_value' is a valid name or a valid value of
        that 'enum' YTAEnum class, or raises an Exception if not.
        """
        return from_name_or_value(name_or_value, cls)
    
    @classmethod
    def to_enum(cls, name_or_value_or_enum):
        """
        This method returns this 'enum' YTAEnum item instance if the
        provided 'name_or_value_or_enum' is a valid name, a valid
        value or a valid instance of this 'enum' YTAEnum class, or
        raises an Exception if not.
        """
        return from_name_or_value_or_enum(name_or_value_or_enum, cls)

    @classmethod
    def get_all(cls):
        """
        This method returns all the existing items in this 'cls' Enum
        class.
        """
        return get_all(cls)
    
    @classmethod
    def get_all_names(cls):
        """
        This method returns all the names of the existing items in
        this 'cls' Enum class.
        """
        return get_all_names(cls)

    @classmethod
    def get_all_values(cls):
        """
        This method returns all the values of the existing items in 
        this 'cls' Enum class.
        """
        return get_all_values(cls)
    
    @classmethod
    def get_all_values_as_str(cls):
        """
        This method returns all the values of the existing items as
        strings separated by commas. This is useful to show the 
        accepted values of an Enum.
        """
        return get_values_as_str(cls.get_all())
    
    @classmethod
    def get_all_names_as_str(cls):
        return get_names_as_str(cls.get_all())





def is_enum(cls: Enum):
    """
    This method returns True if the provided 'cls' parameter is
    an enum class or subclass.
    """
    return isinstance(cls, Enum) or issubclass(cls, (Enum, YTAEnum))

def is_valid(name_or_value_or_enum: any, enum: YTAEnum):
    """
    Returns True if the provided 'name_or_value_or_enum' is
    a valid name or a valid value of the also provided 'enum'
    YTAEnum object, or even if it is an YTAEnum instance of
    that 'enum' YTAEnum class, or False if not.

    This method returns True or False.
    """
    ParameterValidator.validate_mandatory_parameter('name_or_value_or_enum', name_or_value_or_enum)
    ParameterValidator.validate_is_enum(enum)

    return isinstance(name_or_value_or_enum, enum) or is_valid_name(name_or_value_or_enum, enum) or is_valid_value(name_or_value_or_enum)

def is_valid_name_or_value(name_or_value: any, enum: YTAEnum):
    """
    Returns True if the provided 'name_or_value' is a valid
    name or a valid value of the also provided 'enum' YTAEnum 
    object, or False if not.

    This method returns True or False.
    """
    ParameterValidator.validate_mandatory_parameter('name_or_value', name_or_value)
    ParameterValidator.validate_is_enum(enum)

    return is_valid_name(name_or_value) or is_valid_value(name_or_value)

def is_valid_name(name: any, enum: YTAEnum):
    """
    Returns True if the provided 'name' is a valid name of
    the also provided 'enum' YTAEnum object, or False if not.

    This method returns True or False.
    """
    ParameterValidator.validate_mandatory_parameter('name', name)
    ParameterValidator.validate_is_enum(enum)

    try:
        enum[name]

        return True
    except Exception:
        return False
    
def is_valid_value(value: any, enum: YTAEnum):
    """
    Returns True if the provided 'value' is a valid value of
    the also provided 'enum' YTAEnum object.

    This method returns True or False.
    """
    ParameterValidator.validate_mandatory_parameter('value', value)
    ParameterValidator.validate_is_enum(enum)

    try:
        enum(value)

        return True
    except Exception:
        return False
    
def from_name(name: any, enum: YTAEnum):
    """
    This method returns the 'enum' YTAEnum item instance if the
    provided 'name' is a valid name of that 'enum' YTAEnum class,
    or raises an Exception if not.
    """
    if is_valid_name(name, enum):
        return enum[name]
    
    raise Exception(ErrorMessage.parameter_is_not_name_of_ytaenum_class(name, enum))

def from_value(value: any, enum: YTAEnum):
    """
    This method returns the 'enum' YTAEnum item instance if the
    provided 'value' is a valid value of that 'enum' YTAEnum class,
    or raises an Exception if not.
    """
    if is_valid_value(value, enum):
        return enum(value)
    
    raise Exception(ErrorMessage.parameter_is_not_value_of_ytaenum_class(value, enum))

def from_name_or_value(name_or_value: any, enum: YTAEnum):
    """
    This method returns the 'enum' YTAEnum item instance if the
    provided 'name_or_value' is a valid name or a valid value of
    that 'enum' YTAEnum class, or raises an Exception if not.
    """
    if is_valid_name(name_or_value, enum):
        return enum[name_or_value]
    elif is_valid_value(name_or_value, enum):
        return enum(name_or_value)
    
    raise Exception(ErrorMessage.parameter_is_not_name_nor_value_of_ytaenum_class(name_or_value, enum))

def from_name_or_value_or_enum(name_or_value_or_enum: any, enum: YTAEnum):
    """
    This method returns the 'enum' YTAEnum item instance if the
    provided 'name_or_value_or_enum' is a valid name, a valid
    value or a valid instance of that 'enum' YTAEnum class, or
    raises an Exception if not.
    """
    if isinstance(name_or_value_or_enum, enum):
        return enum
    elif is_valid_name(name_or_value_or_enum, enum):
        return enum[name_or_value_or_enum]
    elif is_valid_value(name_or_value_or_enum, enum):
        return enum(name_or_value_or_enum)
    
    raise Exception(ErrorMessage.parameter_is_not_name_nor_value_nor_enum_of_ytaenum_class(name_or_value_or_enum, enum))
    
def is_name_or_value(name_or_value: any, enum: YTAEnum):
    """
    This method validates if the provided 'value' is a name
    or a value of the also provided 'enum' YTAEnum, raising
    an Exception if not.

    This method returns the enum item (containing .name and
    .value) if it is valid.
    """
    ParameterValidator.validate_mandatory_parameter('name_or_value', name_or_value)
    ParameterValidator.validate_is_enum(enum)

    if name_or_value in enum.get_all_names():
        return enum[name_or_value]
    elif name_or_value in enum.get_all_values():
        return enum(name_or_value)
    
    raise Exception(ErrorMessage.parameter_is_not_name_nor_value_of_ytaenum_class(name_or_value, enum))
    
def get_all(enum: Enum):
    """
    This method returns all the items defined in a Enum subtype that
    is provided as the 'enum' parameter.
    """
    if not enum:
        raise Exception('No "enum" provided.')
    
    if not isinstance(enum, Enum) and not issubclass(enum, Enum):
        raise Exception('The "enum" parameter provided is not an Enum.')
    
    return [item for item in enum]

def get_all_names(cls):
    """
    Returns a list containing all the registered enum names.
    """
    return [item.name for item in get_all(cls)]

def get_all_names_as_str(cls):
    """
    Returns a string containing all the enums of the provided
    YTAEnum 'cls' class names separated by commas.
    """
    return get_names_as_str(get_all(cls))

def get_all_values(cls):
    """
    Returns a list containing all the registered enum values.
    """
    return [item.value for item in get_all(cls)]

def get_all_values_as_str(cls):
    """
    Returns a string containing all the enums of the provided
    YTAEnum 'cls' class values separated by commas.
    """
    return get_values_as_str(get_all(cls))

def get_names(enums: list[Enum]):
    """
    Returns a list containing all the names of the provided
    'enums' Enum elements.
    """
    if any(not is_enum(enum) for enum in enums):
        raise TypeError('At least one of the given "enums" is not an Enum class or subclass.')    
    
    return [item.name for item in enums]

def get_names_as_str(enums: list[YTAEnum]):
    """
    Returns a string containing the provided 'enums' names separated
    by commas.
    """
    if any(not is_enum(enum) for enum in enums):
        raise TypeError('At least one of the given "enums" is not an Enum class or subclass.')    
    
    return ', '.join(get_names(enums))

def get_values(enums: list[Enum]):
    """
    Returns a list containing all the values of the provided
    'enums' Enum elements.
    """
    if any(not is_enum(enum) for enum in enums):
        raise TypeError('At least one of the given "enums" is not an Enum class or subclass.')    
     
    return [item.value for item in enums]

def get_values_as_str(enums: Enum):
    """
    Returns a string containing the provided 'enums' values separated
    by commas.
    """
    if any(not is_enum(enum) for enum in enums):
        raise TypeError('At least one of the given "enums" is not an Enum class or subclass.')    
     
    return ', '.join(get_values(enums))